import time
from dataclasses import dataclass, field
from datetime import datetime, timedelta
from operator import add, sub
from types import EllipsisType
from typing import Any, Callable, Dict, Optional, Sequence, TypeVar, overload

from jwt import PyJWT

from .types import AlgorithmType, DateType, RSAPrivateKey, RSAPublicKey


R = TypeVar("R")
DAY = 86400
DEFAULT_EXPIRATION = timedelta(days=31).total_seconds()
NBF_DELTA = 20
ALGORITHMS: Sequence[AlgorithmType] = ("RS256", "RS384", "RS512")


def date_to_timestamp(
    value: DateType | EllipsisType,
    default: Callable[[], R],
    timedelta_func: Callable[[float, float], int] = add,
) -> int | float | R:
    if isinstance(value, timedelta):
        return timedelta_func(time.time(), value.total_seconds())
    elif isinstance(value, datetime):
        return value.timestamp()
    elif isinstance(value, (int, float)):
        return value
    elif value is Ellipsis:
        return default()

    raise ValueError(type(value))


@dataclass(frozen=True, init=False)
class JWTDecoder:
    jwt: PyJWT = field(repr=False, compare=False)
    public_key: RSAPublicKey = field(repr=False, compare=True)
    expires: int | float
    nbf_delta: int | float
    algorithm: AlgorithmType
    algorithms: Sequence[AlgorithmType]

    def __init__(
        self,
        key: RSAPublicKey,
        *, options: dict[str, Any] | None = None,
        expires: int | float = DEFAULT_EXPIRATION,
        nbf_delta: int | float = NBF_DELTA,
        algorithm: AlgorithmType = "RS512",
        algorithms: Sequence[AlgorithmType] = ALGORITHMS,
    ):
        super().__setattr__("public_key", key)
        super().__setattr__("jwt", PyJWT(options))
        super().__setattr__("expires", expires)
        super().__setattr__("nbf_delta", nbf_delta)
        super().__setattr__("algorithm", algorithm)
        super().__setattr__("algorithms", algorithms)

    def decode(self, token: str, verify: bool = True, **kwargs: Any) -> Dict[str, Any]:
        return self.jwt.decode(token, key=self.public_key, verify=verify, algorithms=self.algorithms, **kwargs)


@dataclass(frozen=True, init=False)
class JWTSigner(JWTDecoder):
    private_key: RSAPrivateKey = field(repr=False, compare=True)

    def __init__(self, key: RSAPrivateKey, *, options: Optional[Dict[str, Any]] = None, **kwargs: Any):
        super(JWTDecoder, self).__setattr__("private_key", key)
        super().__init__(key.public_key(), options=options, **kwargs)

    def encode(self, expired: DateType | EllipsisType = ..., nbf: DateType | EllipsisType = ..., **claims: Any) -> str:
        claims.setdefault("exp", int(date_to_timestamp(expired, lambda: time.time() + self.expires)))
        claims.setdefault("nbf", int(date_to_timestamp(nbf, lambda: time.time() - self.nbf_delta, timedelta_func=sub)))
        return self.jwt.encode(claims, self.private_key, algorithm=self.algorithm)


@overload
def JWT(
    key: RSAPrivateKey, *,
    options: dict[str, Any] | None = None,
    expires: int | float = DEFAULT_EXPIRATION,
    nbf_delta: int | float = NBF_DELTA,
    algorithm: AlgorithmType = "RS512",
    algorithms: Sequence[AlgorithmType] = ALGORITHMS,
) -> JWTSigner: ...


@overload
def JWT(
    key: RSAPublicKey, *,
    options: dict[str, Any] | None = None,
    expires: int | float = DEFAULT_EXPIRATION,
    nbf_delta: int | float = NBF_DELTA,
    algorithm: AlgorithmType = "RS512",
    algorithms: Sequence[AlgorithmType] = ALGORITHMS,
) -> JWTDecoder: ...


def JWT(
    key: RSAPrivateKey | RSAPublicKey,
    *,
    options: dict[str, Any] | None = None,
    expires: int | float = DEFAULT_EXPIRATION,
    nbf_delta: int | float = NBF_DELTA,
    algorithm: AlgorithmType = "RS512",
    algorithms: Sequence[AlgorithmType] = ALGORITHMS,
) -> JWTSigner | JWTDecoder:
    kwargs: dict[str, Any] = dict(
        expires=expires,
        nbf_delta=nbf_delta,
        algorithm=algorithm,
        algorithms=algorithms,
        options=options,
    )

    if isinstance(key, RSAPrivateKey):
        return JWTSigner(key, **kwargs)
    elif isinstance(key, RSAPublicKey):
        return JWTDecoder(key, **kwargs)
    else:
        raise TypeError(f"Invalid key type: {type(key)}")



import json
import sys
from types import SimpleNamespace

from .rsa import generate_rsa, load_private_key, load_public_key
from .token import JWT, JWTDecoder, JWTSigner


def main(arguments: SimpleNamespace) -> None:
    jwt: JWTSigner | JWTDecoder
    if arguments.private_key:
        jwt = JWT(load_private_key(arguments.private_key), algorithm=arguments.algorithm)
    elif arguments.public_key:
        jwt = JWT(load_public_key(arguments.public_key), algorithm=arguments.algorithm)
    elif not arguments.verify:
        key_pair = generate_rsa(1024)
        jwt = JWT(key_pair.private, algorithm=arguments.algorithm)
    else:
        print("Either private or public key must be provided", file=sys.stderr)
        exit(1)

    if arguments.interactive:
        token = input("Enter JWT token: ")
    else:
        token = sys.stdin.read()

    print("\nDecoded token:\n")
    print(
        json.dumps(
            jwt.decode(
                token,
                verify=arguments.verify,
                options=dict(
                    verify_signature=arguments.verify,
                    verify_exp=arguments.verify,
                    verify_nbf=arguments.verify,
                    verify_iat=arguments.verify,
                    verify_aud=False,
                ),
            ),
            indent=1,
            sort_keys=True,
        ),
    )



import os
import sys
from types import SimpleNamespace

from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import padding

from .rsa import load_private_key, load_public_key


def main(arguments: SimpleNamespace) -> None:
    public_key = load_public_key(arguments.public)
    private_key = load_private_key(arguments.private)

    payload = os.urandom(1024 * 16)

    signature = private_key.sign(
        payload,
        padding.PSS(
            mgf=padding.MGF1(hashes.SHA256()),
            salt_length=padding.PSS.MAX_LENGTH,
        ),
        hashes.SHA256(),
    )

    print("Signing OK", file=sys.stderr)

    public_key.verify(
        signature,
        payload,
        padding.PSS(mgf=padding.MGF1(hashes.SHA256()), salt_length=padding.PSS.MAX_LENGTH),
        hashes.SHA256(),
    )

    print("Verifying OK", file=sys.stderr)



import sys
from pathlib import Path
from types import SimpleNamespace

from .convert import convert
from .rsa import generate_rsa


def main(arguments: SimpleNamespace) -> None:
    key_pair = generate_rsa(arguments.bits)

    if not arguments.force:
        if arguments.save_public and arguments.save_public.exists():
            print("Public key file already exists, use --force to overwrite. Not saving.", file=sys.stderr)
            arguments.save_public = None

        if arguments.save_private and arguments.save_private.exists():
            print("Private key file already exists, use --force to overwrite. Not saving.", file=sys.stderr)
            arguments.save_private = None

    if isinstance(arguments.save_private, Path) and not arguments.save_public:
        save_public = arguments.save_private.with_name(arguments.save_private.stem + ".pub")
        if arguments.force or not save_public.exists():
            arguments.save_public = save_public
            print("Public key file not specified, saving public key to", save_public, file=sys.stderr)

    public_data, private_data = convert(
        key_pair.private,
        key_pair.public,
        arguments.format,
        arguments.pretty,
    )

    if arguments.save_public:
        print("Saving public key to", arguments.save_public, "in PEM format", file=sys.stderr)
        with open(arguments.save_public, "w") as fp:
            fp.write(public_data)
    else:
        print("Public key in", arguments.format, "format:", file=sys.stderr)
        print(public_data)

    if arguments.save_private:
        print("Saving private key to", arguments.save_private, "in PEM format", file=sys.stderr)
        with open(arguments.save_private, "w") as fp:
            fp.write(private_data)
    else:
        print("Private key in", arguments.format, "format:", file=sys.stderr)
        print(private_data)



import base64
import hashlib
import json
import sys
from pathlib import Path
from types import SimpleNamespace
from typing import Literal

from .rsa import load_private_key, rsa_to_jwk
from .types import AlgorithmType, RSAPrivateKey, RSAPublicKey, serialization


def generate_kid(key: RSAPrivateKey) -> str:
    public_numbers = key.public_key().public_numbers()
    modulus_bytes = public_numbers.n.to_bytes((public_numbers.n.bit_length() + 7) // 8, byteorder="big")
    hash_object = hashlib.sha256(modulus_bytes)
    kid = base64.urlsafe_b64encode(hash_object.digest()).decode("utf-8").rstrip("=")
    return kid[:16]


def convert(
    private: RSAPrivateKey, public: RSAPublicKey,
    fmt: Literal["pem", "jwk", "base64"],
    pretty: bool = False, algorithm: AlgorithmType = "RS512",
) -> tuple[str, str]:
    if fmt == "pem":
        return (
            public.public_bytes(
                encoding=serialization.Encoding.PEM,
                format=serialization.PublicFormat.PKCS1,
            ).decode(),
            private.private_bytes(
                encoding=serialization.Encoding.PEM,
                encryption_algorithm=serialization.NoEncryption(),
                format=serialization.PrivateFormat.PKCS8,
            ).decode(),
        )

    elif fmt == "jwk":
        kid = generate_kid(private)
        public_jwk = rsa_to_jwk(public, kid=kid, alg=algorithm)
        private_jwk = rsa_to_jwk(private, kid=kid, alg=algorithm)

        return (
            json.dumps(public_jwk, sort_keys=True, indent=" " if pretty else None),
            json.dumps(private_jwk, sort_keys=True, indent=" " if pretty else None),
        )
    elif fmt == "base64":
        return (
            base64.b64encode(
                public.public_bytes(
                    encoding=serialization.Encoding.DER,
                    format=serialization.PublicFormat.PKCS1,
                ),
            ).decode(),
            base64.b64encode(
                private.private_bytes(
                    encoding=serialization.Encoding.DER,
                    format=serialization.PrivateFormat.PKCS8,
                    encryption_algorithm=serialization.NoEncryption(),
                ),
            ).decode(),
        )
    raise NotImplementedError("Unknown format", format)


def main(arguments: SimpleNamespace) -> None:
    private = load_private_key(arguments.private_key)
    public = private.public_key()

    if not arguments.force:
        if arguments.save_public and arguments.save_public.exists():
            print("Public key file already exists, use --force to overwrite. Not saving.", file=sys.stderr)
            arguments.save_public = None

        if arguments.save_private and arguments.save_private.exists():
            print("Private key file already exists, use --force to overwrite. Not saving.", file=sys.stderr)
            arguments.save_private = None

    if isinstance(arguments.save_private, Path) and not arguments.save_public:
        save_public = arguments.save_private.with_name(arguments.save_private.stem + ".pub")
        if arguments.force or not save_public.exists():
            arguments.save_public = save_public
            print("Public key file not specified, saving public key to", save_public, file=sys.stderr)

    public_data, private_data = convert(
        private,
        public,
        arguments.format,
        arguments.pretty,
    )

    if arguments.save_public:
        print("Saving public key to", arguments.save_public, "in PEM format", file=sys.stderr)
        with open(arguments.save_public, "w") as fp:
            fp.write(public_data)
    else:
        print("Public key in", arguments.format, "format:", file=sys.stderr)
        print(public_data)

    if arguments.save_private:
        print("Saving private key to", arguments.save_private, "in PEM format", file=sys.stderr)
        with open(arguments.save_private, "w") as fp:
            fp.write(private_data)
    else:
        print("Private key in", arguments.format, "format:", file=sys.stderr)
        print(private_data)
